#ifndef FrameCPP__COMMON__STRING_HH
#define FrameCPP__COMMON__STRING_HH

#include <stdexcept>
#include <string>
#include <vector>

#include "general/types.hh"
#include "general/autoarray.hh"

#include "framecpp/Common/IOStream.hh"

namespace FrameCPP
{
  namespace Common
  {
    template < class L >
    class STRING
      : public std::string
    {
    public:
      typedef L length_type;

      static const length_type MAX_STRING_LENGTH;

      STRING( );

      STRING( const std::string& Value );

      STRING( const char* Value );

      //: Calculate the number of bytes needed to write this object
      INT_8U Bytes( ) const;

      static INT_8U Bytes( const std::string& String );

      STRING& operator=( const std::string& Source );

      STRING& operator=( const char* Source );

      void Read( IStream& Stream );

      void Write( OStream& Stream ) const;

    private:
      static void	length_error( const std::string& Value );
      static void	length_error( const char* Value );

    }; // class - STRING

    template< class L >
    const L STRING< L >::MAX_STRING_LENGTH =
      ~( L( 0 ) );

    template< class L >
    STRING< L >::
    STRING( )
      : std::string( )
    {
    }

    template< class L >
    STRING< L >::
    STRING( const std::string& Value )
    {
      length_error( Value );
      std::string::operator=( Value );
    }

    template< class L >
    STRING< L >::
    STRING( const char* Value )
    {
      length_error( Value );
      std::string::operator=( Value );
    }

    template< class L >
    STRING< L >& STRING< L >::
    operator=( const std::string& Source )
    {
      length_error( Source );
      std::string::operator=( Source );
      return *this;
    }

    template< class L >
    STRING< L >& STRING< L >::
    operator=( const char* Source )
    {
      length_error( Source );
      std::string::operator=( Source );
      return *this;
    }

    template< class L >
    INT_8U STRING< L >::
    Bytes( ) const
    {
      return length( )		 // Length of string
	+ 1			 // space for the null character
	+ sizeof( length_type ); // length to hold the size
    }

    template< class L >
    INT_8U STRING< L >::
    Bytes( const std::string& String )
    {
      return String.length( )	 // Length of string
	+ 1			 // space for the null character
	+ sizeof( length_type ); // length to hold the size
    }

    template< class L >
    void STRING< L >::
    Read( IStream& Stream )
    {
      length_type	l = 0;
      Stream >> l;
      if ( l > 0 )
      {
	General::AutoArray< char >	c( new char[ l ] );

	Stream.read( c.get( ), l );
	assign( c.get( ) );
      }
      else
      {
	assign( "" );
      }
    }

    template< class L >
    void STRING< L >::
    Write( OStream& Stream ) const
    {
      length_type	l( length( ) + 1 );
      Stream << l;
      Stream.write( c_str( ), l );
    }

    template< class L >
    inline void STRING< L >::
    length_error( const std::string& Buffer )
    {
      if ( Buffer.length( ) > MAX_STRING_LENGTH )
      {
	std::ostringstream msg;
      
	msg << "maximum size of STRING exceeded: "
	    << Buffer.length( ) << " > " << MAX_STRING_LENGTH
	  ;
	throw std::length_error( msg.str( ) );
      }
    }

    template< class L >
    inline void STRING< L >::
    length_error( const char* Buffer )
    {
      if ( strlen( Buffer ) > MAX_STRING_LENGTH )
      {
	std::ostringstream msg;
      
	msg << "maximum size of STRING exceeded: "
	    << strlen( Buffer ) << " > " << MAX_STRING_LENGTH
	  ;
	throw std::length_error( msg.str( ) );
      }
    }

    template< class L >
    inline IStream&
    operator>>( IStream& Stream,
		STRING< L >& S )
    {
      S.Read( Stream );
      return Stream;
    }

    template< typename L >
    inline IStream&
    operator>>( IStream& Stream, std::vector< STRING< L > >& Data )
    {
      for( typename std::vector< STRING< L > >::iterator
	     cur = Data.begin( ),
	     last = Data.end( );
	   cur != last;
	   ++cur )
      {
	Stream >> *cur;
      }
      return Stream;
    }

    template< class L >
    inline OStream&
    operator<<( OStream& Stream,
		const STRING< L >& S )
    {
      S.Write( Stream );
      return Stream;
    }

    template< class L >
    inline OStream&
    operator<<( OStream& Stream,
		const std::vector< STRING< L > >& Data )
    {
      for ( typename std::vector< STRING< L > >::const_iterator
	      cur = Data.begin( ),
	      last = Data.end( );
	    cur != last;
	    ++cur )
      {
	Stream << *cur;
      }
      return Stream;
    }

    inline void
    AppendComment( std::string& Dest, const std::string& Text )
    {
      if( Dest.size() )
      {
	Dest += '\n' + Text;
      }
      else
      {
	Dest = Text;
      }
    }
  } // namespace - Common

} // namespace - FrameCPP

#endif /* FrameCPP__COMMON__STRING_HH */
