/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef GENERAL__GPS_TIME_HH
#define GENERAL__GPS_TIME_HH

#include <stdexcept>
#include <iostream>

#include "general/ErrorLog.hh"
#include "general/types.hh"

const REAL_8 NANOSECOND( 10E-9 );
const REAL_8 NANOSECOND_MULTIPLIER( 10E+9 );
const INT_4U NANOSECOND_INT_MULTIPLIER( 1000000000 );

namespace General {

    class GPSTime {

    public:
	typedef enum {
	    UTC = 0,
	    GPS
	} unit_type;

	static const unit_type DEFAULT_TIME_UNIT_TYPE;

        GPSTime();
        GPSTime( const INT_4U seconds,
		 const INT_4U nanoseconds,
		 unit_type Units = DEFAULT_TIME_UNIT_TYPE );

        GPSTime(const GPSTime&);

        ~GPSTime();

        GPSTime& operator=(const GPSTime&);

        INT_4U GetSeconds( ) const;
        INT_4U GetSeconds( const unit_type Units ) const;
        INT_4U GetNanoseconds() const;
	REAL_8 GetTime( ) const;
        INT_2U GetLeapSeconds() const;

        GPSTime& operator+=(const double&);
        GPSTime& operator-=(const double&);

	/// \brief  Sets the object to the current GPS time.
	void Now();

	/// Helper function to make it compatable with Time class of
	///   framecpp
	INT_4U getSec( ) const;

	/// Helper function to make it compatable with Time class of
	///   framecpp
	INT_4U getNSec( ) const;
        
        /// \brief  Returns the time "now" as a GPSTime
        /// \return GPSTime the current time, in GPSTime form
        static GPSTime NowGPSTime( );

    protected:
        /// \brief  Returns the time "now" as a GPSTime
        /// \return GPSTime the current time, in GPSTime form
        static GPSTime now( );

    private:
	friend void ErrorLog::operator()( state State,
					  const char* Filename, const int Line,
					  const std::string& Message,
					  bool EOL );
        INT_4U m_seconds;
        INT_4U m_nanoseconds;

    }; // class GPSTime

    inline REAL_8 GPSTime::
    GetTime( ) const
    {
	return ( m_seconds + ( m_nanoseconds / NANOSECOND_MULTIPLIER ) );
    }

    inline INT_4U GPSTime::
    GetNanoseconds() const
    {
	return m_nanoseconds;
    }

    inline INT_4U GPSTime::
    GetSeconds( ) const
    {
	return m_seconds;
    }

    inline INT_4U GPSTime::
    getSec( ) const
    {
	return GetSeconds( );
    }

    /// Helper function to make it compatable with Time class of
    ///   framecpp
    inline INT_4U GPSTime::
    getNSec( ) const
    {
	return GetNanoseconds( );
    }

    //-------------------------------------------------------------------
    /// Establish the filename which contains the leap seconds
    //-------------------------------------------------------------------
    void SetOffsetTableFilename( const std::string& Filename );

    GPSTime operator+(const GPSTime&, const double&);
    GPSTime operator+(const double&, const GPSTime&);

    GPSTime operator-(const GPSTime&, const double&);
    double operator-(const GPSTime&, const GPSTime&);

    bool operator==(const GPSTime&, const GPSTime&);
    bool operator!=(const GPSTime&, const GPSTime&);

    bool operator<(const GPSTime&, const GPSTime&);
    bool operator>(const GPSTime&, const GPSTime&);

    bool operator<=(const GPSTime&, const GPSTime&);
    bool operator>=(const GPSTime&, const GPSTime&);
    
    std::ostream& operator<<( std::ostream& Stream,
                              const GPSTime& Time );

} // namespace General

#endif // GENERAL__GPS_TIME_HH
